"""
Modern rewrite of IFeelBloated's package with scene detection.
"""
from __future__ import annotations

from functools import partial
from math import log
from typing import Any

from vsdenoise import CCDMode, CCDPoints, ChannelMode, MVTools, PelType, Prefilter, ccd, knl_means_cl
from vsexprtools import norm_expr_planes
from vskernels import Bicubic
from vsrgtools import contrasharpening_dehalo, gauss_blur, gauss_fmtc_blur, lehmer_diff_merge
from vstools import (
    Matrix, PlanesT, check_ref_clip, core, disallow_variable_format, disallow_variable_resolution, get_y, join,
    normalize_planes, split, vs
)

from . import masks
from .alpha import fine_dehalo

__all__ = [
    'super_clip', 'smooth_clip', 'dehalo'
]


@disallow_variable_format
@disallow_variable_resolution
def super_clip(src: vs.VideoNode, pel: int = 1, planes: PlanesT = 0) -> vs.VideoNode:
    assert src.format

    if planes == [1, 2] and (src.format.subsampling_w or src.format.subsampling_h):
        src = src.resize.Bicubic(
            src.width // (1 << src.format.subsampling_w),
            src.height // (1 << src.format.subsampling_h),
            format=src.format.replace(subsampling_w=0, subsampling_h=0).id
        )
    elif planes == [0]:
        src = get_y(src)

    return PelType.NNEDI3(src, pel, nns=4, qual=2, etype=1)


@disallow_variable_format
@disallow_variable_resolution
def smooth_clip(
    src: vs.VideoNode, sr: int = 32, strength: float = 12.5,
    sharp: float = 0.80, cutoff: int = 4,
    aggressive: bool = True, fast: bool = False,
    matrix: int | Matrix | None = None,
    pel_type: PelType = PelType.BICUBIC,
    planes: PlanesT = 0
) -> vs.VideoNode:
    assert src.format

    if sr < 1:
        raise RuntimeError('vine.smooth_clip: sr has to be greater than 0!')

    if strength <= 0:
        raise RuntimeError('vine.smooth_clip: strength has to be greater than 0!')

    if sharp <= 0.0:
        raise RuntimeError('vine.smooth_clip: sharp has to be greater than 0!')

    if cutoff < 1 or cutoff > 100:
        raise RuntimeError('vine.smooth_clip: cutoff must fall in (0, 100]!')

    csp = src.format.color_family
    planes = normalize_planes(src, planes)

    if fast and csp == vs.GRAY:
        raise ValueError('vine.smooth_clip: fast=True is available only for YUV and RGB input!')

    work_clip = src

    if csp == vs.RGB:
        work_clip = work_clip.bm3d.RGB2OPP(1)

    work_clip, *chroma = split(work_clip) if planes == [0] and not fast else (work_clip, )
    assert work_clip.format

    c1 = 0.3926327792690057290863679493724 * sharp
    c2 = 18.880334973195822973214959957208
    c3 = 0.5862453661304626725671053478676

    weight = c1 * log(1.0 + 1.0 / (c1))

    h_smoothine = c2 * pow(strength / c2, c3)

    upsampled = pel_type(work_clip, 2)

    knl_channels = ChannelMode.from_planes(planes)

    if fast:
        upsampled = ccd(
            upsampled, strength * 2, 0, None,
            CCDMode.BICUBIC_LUMA, 25 / (2 * sr + 1), matrix,
            CCDPoints.MEDIUM, False, planes
        )
        upsampled = gauss_blur(upsampled, 1.45454, planes=planes)
    else:
        upsampled = knl_means_cl(upsampled, strength, 0, sr, 0, knl_channels)

    c = 0 if pel_type == PelType.NNEDI3 else (1.0 - abs(sharp)) / 2

    resampled = Bicubic(b=-sharp, c=c).scale(upsampled, work_clip.width, work_clip.height)

    if fast:
        resampled = contrasharpening_dehalo(resampled, work_clip, 2.5, planes=planes)

    clean = knl_means_cl(work_clip, strength, 0, sr, 0, knl_channels)

    clean = resampled.std.Merge(
        clean, [weight if i in planes else 0 for i in range(work_clip.format.num_planes)]
    )

    blur_func = partial(
        gauss_fmtc_blur, sigma=cutoff, sharp=100, strict=False, planes=planes
    )

    if aggressive:
        clean = core.std.Expr(
            [blur_func(work_clip), clean, blur_func(clean)],
            norm_expr_planes(work_clip, 'y z - x +', planes)
        )
    else:
        clean = lehmer_diff_merge(
            gauss_blur(work_clip, 0.5), clean, blur_func, planes=planes
        )

    diff = work_clip.std.MakeDiff(clean, planes)

    diff = knl_means_cl(diff, h_smoothine, 0, sr, 1, knl_channels, rclip=clean)

    smooth = clean.std.MergeDiff(diff, planes)

    if chroma:
        smooth = join([smooth, *chroma], vs.YUV)

    if csp == vs.RGB:
        return smooth.bm3d.OPP2RGB(1)

    return smooth


@disallow_variable_format
@disallow_variable_resolution
def dehalo(
    src: vs.VideoNode, smooth: vs.VideoNode | None = None,
    tr: int = 0, refine: int = 3, pel: int = 1, thSAD: int = 400,
    super_clips: tuple[vs.VideoNode | None, vs.VideoNode | None] = (None, None),
    planes: PlanesT = 0, mask: bool | vs.VideoNode = True
) -> vs.VideoNode:
    assert src.format

    if smooth:
        assert smooth.format

        if src.format.id != smooth.format.id or (src.width, src.height) != (smooth.width, smooth.height):
            raise TypeError('vine.dehalo: smooth must have the same format and size as src!')

    if pel not in {1, 2, 4}:
        raise RuntimeError('vine.dehalo: pel has to be 1, 2 or 4!')

    if thSAD <= 0:
        raise RuntimeError('vine.dehalo: sad has to be greater than 0!')

    csp = src.format.color_family
    planes = normalize_planes(src, planes)

    if any(sclip and sclip.format and sclip.format.num_planes == 1 for sclip in super_clips):
        planes = [0]

    if smooth is None:
        smooth = smooth_clip(src, planes=planes)

    work_clip, smooth_wclip = src, smooth

    if csp == vs.RGB:
        work_clip = work_clip.bm3d.RGB2OPP(1)
        smooth_wclip = smooth_wclip.bm3d.RGB2OPP(1)

    work_clip, *chroma = split(work_clip) if planes == [0] else (work_clip, )
    smooth_wclip = get_y(smooth_wclip) if planes == [0] else smooth_wclip

    constant = 0.0009948813682897925944723492342

    me_sad = constant * pow(thSAD, 2.0) * log(1.0 + 1.0 / (constant * thSAD))

    if isinstance(mask, vs.VideoNode):
        check_ref_clip(src, (halo_mask := mask))
    elif mask:
        halo_mask = fine_dehalo(
            work_clip, None, 2.1, 2.1, ss=1, edgeproc=0.5, show_mask=True, planes=planes
        )
        halo_mask = masks.dilation(halo_mask, 1, planes)
    else:
        halo_mask = None

    smooth_wclip = smooth_wclip.std.MakeDiff(work_clip, planes)

    src_super, smooth_super = super_clips

    if pel == 1:
        if not src_super:
            src_super = work_clip
        if not smooth_super:
            smooth_super = smooth_wclip

    if src_super and smooth_super:
        smooth_super = smooth_super.std.MakeDiff(src_super)

    mv = MVTools(
        smooth_wclip, tr, refine, prefilter=Prefilter.NONE,
        pel=pel, rfilter=4, planes=planes, fix_fades=False
    )
    mv.subpel_clips = (smooth_super, src_super)
    mv.analyze_args |= dict[str, Any](trymany=True, badrange=-24, divide=0)
    mv.recalculate_args |= dict[str, Any](smooth=1, divide=0, thsad=me_sad)

    mv.analyze(work_clip, search=3, truemotion=False)

    averaged_dif = mv.degrain(thSAD=thSAD)

    averaged_dif = core.std.Expr(
        [averaged_dif, smooth_wclip], norm_expr_planes(work_clip, 'x y min', planes)
    )

    clean = work_clip.std.MergeDiff(averaged_dif, planes)

    if halo_mask:
        clean = work_clip.std.MaskedMerge(clean, halo_mask, planes)

    if chroma:
        clean = join([clean, *chroma], vs.YUV)

    if csp == vs.RGB:
        return clean.bm3d.OPP2RGB(1)

    return clean



from __future__ import annotations

from functools import partial
from math import ceil
from typing import Any, Literal

from vsdenoise import BM3D, BM3DCPU, BM3DCuda, BM3DCudaRTC, Prefilter
from vsexprtools import norm_expr_planes
from vsmask.edge import Prewitt
from vsmask.util import expand, inpand
from vsrgtools import LimitFilterMode, contrasharpening, contrasharpening_dehalo, limit_filter, repair
from vstools import (
    PlanesT, core, depth, disallow_variable_format, disallow_variable_resolution, fallback, get_depth, get_peak_value,
    get_y, iterate, join, normalize_planes, normalize_seq, scale_value, split, vs
)

__all__ = [
    'bidehalo', 'HQDeringmod'
]


@disallow_variable_format
@disallow_variable_resolution
def bidehalo(
    clip: vs.VideoNode,
    sigma: float = 1.5, radius: float = 7,
    sigma_final: float | None = None, radius_final: float | None = None,
    tr: int = 2, cuda: bool | Literal['rtc'] = False,
    planes: PlanesT = 0, matrix: int | vs.MatrixCoefficients | None = None,
    bm3d_args: dict[str, Any] | None = None, bilateral_args: dict[str, Any] | None = None
) -> vs.VideoNode:
    """
    Simple dehalo function that uses ``bilateral`` and ``BM3D`` to remove bright haloing around edges.

    This works by utilising the ``ref`` parameter in ``bilateral`` to limit the areas that get damaged,
    and how much it gets damaged. You should use this function in conjunction with a halo mask.

    If a ref clip is passed, that will be used as a ref for the second bilateral pass instead of a blurred clip.
    Both clips will be resampled to 16bit internally, and returned in the input bitdepth.

    Recommend values for `sigma` are between 0.8 and 2.0.
    Recommend values for `radius` are between 5 and 15.

    Dependencies:

    * VapourSynth-Bilateral (Default)
    * VapourSynth-BilateralGPU (Cuda)
    * VapourSynth-BilateralGPU_RTC (RTC)
    * vsdenoise

    :param clip:                Clip to process.
    :param sigma:               ``Bilateral`` spatial weight sigma.
    :param sigma_final:         Final ``Bilateral`` call's spatial weight sigma.
                                You'll want this to be much weaker than the initial `sigma`.
                                If `None`, 1/3rd of `sigma`.
    :param radius:              ``Bilateral`` radius weight sigma.
    :param radius_final:        Final ``Bilateral`` radius weight sigma.
                                if `None`, same as `radius`.
    :param tr:                  Temporal radius for BM3D
    :param cuda:                Use ``BM3DCUDA`` and `BilateralGPU` if True, else ``BM3DCPU`` and `Bilateral`.
                                Also accepts 'rtc' for ``BM3DRTC`` and `BilateralGPU_RTC`.
                                Notice: final pass of bilateral will always be on cpu since the
                                gpu implementation doesn't support passing ``ref``.
    :param planes:              Specifies which planes will be processed.
                                Any unprocessed planes will be simply copied.
    :param bm3d_args:           Additional parameters to pass to BM3D.
    :param bilateral_args:      Additional parameters to pass to Bilateral.

    :return:                    Dehalo'd clip using ``BM3D`` and ``Bilateral``.
    """
    assert clip.format

    bits = get_depth(clip)

    bm3d_args = fallback(bm3d_args, dict[str, Any]())
    bilateral_args = fallback(bilateral_args, dict[str, Any]())

    sigma_final = fallback(sigma_final, sigma / 3)
    radius_final = fallback(radius_final, radius)

    planes = normalize_planes(clip, planes)

    if matrix:
        clip = clip.std.SetFrameProp('_Matrix', int(matrix))

    process_chroma = 1 in planes or 2 in planes

    if not cuda:
        sigma_luma, sigma_chroma = 8, process_chroma and 6.4
    else:
        sigma_luma, sigma_chroma = 10, process_chroma and 8

    bm3d_pargs = (depth(clip, 16), [sigma_luma, sigma_chroma], tr)

    if cuda is False:
        try:
            den = BM3D(*bm3d_pargs, **bm3d_args).clip
        except AttributeError:
            den = BM3DCPU(*bm3d_pargs, **bm3d_args).clip

        ref = den.bilateral.Bilateral(None, sigma, radius / 255, planes, **bilateral_args)
    else:
        bil_gpu_args = dict[str, Any](sigma_spatial=sigma, **bilateral_args)

        if cuda is True:
            den = BM3DCuda(*bm3d_pargs, **bm3d_args).clip
            ref = den.bilateralgpu.Bilateral(**bil_gpu_args)
        elif cuda == 'rtc':
            den = BM3DCudaRTC(*bm3d_pargs, **bm3d_args).clip
            ref = den.bilateralgpu_rtc.Bilateral(**bil_gpu_args)
        else:
            raise ValueError(f'bidehalo: Invalid cuda selection ({cuda})!')

    bidh = den.bilateral.Bilateral(ref, sigma_final, radius_final / 255, planes, **bilateral_args)
    bidh = depth(bidh, bits)

    return core.std.Expr([clip, bidh], norm_expr_planes(clip, 'x y min', planes))


@disallow_variable_format
@disallow_variable_resolution
def HQDeringmod(
    clip: vs.VideoNode,
    smooth: vs.VideoNode | Prefilter | tuple[Prefilter, Prefilter] = Prefilter.MINBLUR1,
    ringmask: vs.VideoNode | None = None,
    mrad: int = 1, msmooth: int = 1, minp: int = 1, mthr: int = 60, incedge: bool = False,
    thr: int = 12, elast: float = 2.0, darkthr: int | None = None,
    sigma: float = 128.0, sigma2: float | None = None,
    sbsize: int | None = None, sosize: int | None = None,
    contra: int | float | bool = 1.2, drrep: int = 13,
    planes: PlanesT = 0, show: bool = False, **kwargs: Any
) -> vs.VideoNode:
    """
    :param clip:        Clip to process.
    :param smooth:      Already smoothed clip, or a Prefilter, tuple for [luma, chroma] prefilter.
    :param ringmask:    Custom ringing mask.
    :param mrad:        Expanding iterations of edge mask, higher value means more aggressive processing.
    :param msmooth:     Inflating iterations of edge mask, higher value means smoother edges of mask.
    :param minp:        Inpanding iterations of prewitt edge mask, higher value means more aggressive processing.
    :param mthr:        Threshold of prewitt edge mask, lower value means more aggressive processing
                        but for strong ringing, lower value will treat some ringing as edge,
                        which "protects" this ringing from being processed.
    :param incedge:     Whether to include edge in ring mask, by default ring mask only include area near edges.
    :param thr:         Threshold (8-bit scale) to limit filtering diff.
                        Smaller thr will result in more pixels being taken from processed clip.
                        Larger thr will result in less pixels being taken from input clip.
                            PDiff: pixel value diff between processed clip and input clip
                            ODiff: pixel value diff between output clip and input clip
                            PDiff, thr and elast is used to calculate ODiff:
                            ODiff = PDiff when [PDiff <= thr]
                            ODiff gradually smooths from thr to 0 when [thr <= PDiff <= thr * elast]
                            For elast>2.0, ODiff reaches maximum when [PDiff == thr * elast / 2]
                            ODiff = 0 when [PDiff >= thr * elast]
    :param elast:       Elasticity of the soft threshold.
                        Larger "elast" will result in more pixels being blended from.
    :param darkthr:     Threshold (8-bit scale) for darker area near edges, for filtering diff
                        that brightening the image by default equals to thr/4.
                        Set it lower if you think de-ringing destroys too much lines, etc.
                        When darkthr is not equal to ``thr``, ``thr`` limits darkening,
                        while ``darkthr`` limits brightening.
                        This is useful to limit the overshoot/undershoot/blurring introduced in deringing.
                        Examples:
                            ``thr=0``,   ``darkthr=0``  : no limiting
                            ``thr=255``, ``darkthr=255``: no limiting
                            ``thr=8``,   ``darkthr=2``  : limit darkening with 8, brightening is limited to 2
                            ``thr=8``,   ``darkthr=0``  : limit darkening with 8, brightening is limited to 0
                            ``thr=255``, ``darkthr=0``  : limit darkening with 255, brightening is limited to 0
                            For the last two examples, output will remain unchanged. (0/255: no limiting)
    :param sigma:       DFTTest Prefilter only: sigma for medium frequecies
    :param sigma2:      DFTTest Prefilter only: sigma for low&high frequecies
    :param sbsize:      DFTTest Prefilter only: length of the sides of the spatial window
    :param sosize:      DFTTest Prefilter only: spatial overlap amount
    :param contra:      Whether to use contra-sharpening to resharp deringed clip:
                            False: no contrasharpening
                            True: auto radius for contrasharpening
                            int 1-3: represents radius for contrasharpening
                            float: represents level for contrasharpening_dehalo
    :param drrep:       Use repair for details retention, recommended values are 13/12/1.
    :param planes:      Planes to be processed.
    :param show:        Show the computed ringing mask.
    :param kwargs:      Kwargs to be passed to the prefilter function.

    :return:            Deringed clip.
    """
    assert clip.format

    if clip.format.color_family not in {vs.YUV, vs.GRAY}:
        raise ValueError('HQDeringmod: format not supported')

    peak = get_peak_value(clip)
    bits = clip.format.bits_per_sample
    planes = normalize_planes(clip, planes)
    work_clip, *chroma = split(clip) if planes == [0] else (clip, )
    assert work_clip.format

    is_HD = clip.width >= 1280 or clip.height >= 720

    # Parameters for deringing kernel
    sigma2 = fallback(sigma2, sigma / 16)
    sbsize = fallback(sbsize, 8 if is_HD else 6)
    sosize = fallback(sosize, 6 if is_HD else 4)

    darkthr = fallback(darkthr, thr // 4)

    rep_dr = [drrep if i in planes else 0 for i in range(work_clip.format.num_planes)]

    # Kernel: Smoothing
    if not isinstance(smooth, vs.VideoNode):
        smoothy, smoothc = normalize_seq(smooth, 2)

        def _get_kwargs(pref: Prefilter) -> dict[str, Any]:
            if pref != Prefilter.DFTTEST:
                return kwargs
            return kwargs | dict(
                sbsize=sbsize, sosize=sosize, tbsize=1, slocation=[
                    0.0, sigma2, 0.05, sigma, 0.5, sigma, 0.75, sigma2, 1.0, 0.0
                ])

        if smoothy == smoothc or work_clip.format.num_planes == 1:
            smoothed = smoothy(work_clip, planes, **_get_kwargs(smoothy))
        else:
            smoothed = core.std.ShufflePlanes([
                smoothy(work_clip, 0, **_get_kwargs(smoothy)),
                smoothc(work_clip, list(set(planes) - {0}), **_get_kwargs(smoothc))
            ], [0, 1, 2], vs.YUV)
    else:
        assert work_clip.format and smooth.format

        if smooth.format.id != clip.format.id:
            raise ValueError("Smooth clip format must match the source clip's!")

        if smooth.width != clip.width or smooth.height != clip.height:
            raise ValueError("Smooth clip sizes must match the source clip's!")

        smoothed = get_y(smooth) if planes == [0] else smooth

    # Post-Process: Contra-Sharpening
    if contra:
        if isinstance(contra, int):
            smoothed = contrasharpening(smoothed, work_clip, contra, 13, planes)
        else:
            smoothed = contrasharpening_dehalo(smoothed, work_clip, contra, planes=planes)

    # Post-Process: Repairing
    if set(rep_dr) != {0}:
        repclp = repair(work_clip, smoothed, drrep)
    else:
        repclp = work_clip

    # Post-Process: Limiting
    limitclp = limit_filter(
        repclp, work_clip, None, LimitFilterMode.CLAMPING, planes, thr, elast, darkthr
    )

    if ringmask is None:
        # FIXME: <= instead of < for lthr, Vardë?
        prewittm = Prewitt().edgemask(work_clip, scale_value(mthr, 8, bits) + 1)

        fmask = prewittm.std.Median(planes).misc.Hysteresis(prewittm, planes)

        omask = expand(fmask, mrad, mrad, planes=planes) if mrad > 0 else fmask

        if msmooth > 0:
            omask = iterate(omask, partial(core.std.Inflate, planes=planes), msmooth)

        if incedge:
            ringmask = omask
        else:
            if minp <= 0:
                imask = fmask
            elif minp % 2 == 0:
                imask = inpand(fmask, minp // 2, planes=planes)
            else:
                imask = inpand(fmask.std.Inflate(planes), ceil(minp / 2), planes=planes)

            ringmask = core.std.Expr([omask, imask], f'x {peak} y - * {peak} /')

    dering = work_clip.std.MaskedMerge(limitclp, ringmask, planes)

    if show:
        return ringmask

    if chroma:
        return join([dering, *chroma], clip.format.color_family)

    return dering



from __future__ import annotations

from vsaa import Nnedi3
from vsexprtools import ExprOp, aka_expr_available, combine, norm_expr
from vskernels import BSpline, Lanczos, Mitchell, NoShift, Point, Scaler, ScalerT
from vsmask.edge import EdgeDetect, Robinson3
from vsmask.util import XxpandMode, expand, inpand
from vsrgtools import box_blur, contrasharpening, contrasharpening_dehalo, repair
from vsrgtools.util import norm_rmode_planes
from vstools import (
    ColorRange, ConvMode, CustomIndexError, CustomValueError, FuncExceptT, InvalidColorFamilyError, PlanesT,
    check_variable, clamp, cround, disallow_variable_format, disallow_variable_resolution, fallback, get_peak_value,
    join, mod4, normalize_planes, normalize_seq, scale_value, split, to_arr, vs
)

from . import masks

__all__ = [
    'fine_dehalo',
    'fine_dehalo2',
    'dehalo_alpha'
]


FloatIterArr = float | list[float] | tuple[float | list[float], ...]


@disallow_variable_format
@disallow_variable_resolution
def fine_dehalo(
    clip: vs.VideoNode, rx: FloatIterArr = 2.0, ry: FloatIterArr | None = None, darkstr: FloatIterArr = 0.0,
    brightstr: FloatIterArr = 1.0, lowsens: FloatIterArr = 50.0, highsens: FloatIterArr = 50.0,
    thmi: int = 80, thma: int = 128, thlimi: int = 50, thlima: int = 100, sigma_mask: float = 0.0,
    ss: FloatIterArr = 1.5, contra: int | float | bool = 0.0, exclude: bool = True,
    edgeproc: float = 0.0, edgemask: EdgeDetect = Robinson3(), planes: PlanesT = 0, show_mask: int | bool = False,
    mask_radius: int = 1, downscaler: ScalerT = Mitchell, upscaler: ScalerT = BSpline,
    supersampler: ScalerT = Lanczos(3), supersampler_ref: ScalerT = Mitchell, pre_ss: float = 1.0,
    pre_supersampler: ScalerT = Nnedi3(0, field=0, shifter=NoShift), pre_downscaler: ScalerT = Point,
    func: FuncExceptT | None = None
) -> vs.VideoNode:
    """
    Halo removal script that uses ``dehalo_alpha`` with a few masks and optional contra-sharpening
    to try removing halos without nuking important details like line edges.

    **For ``rx``, ``ry``, only the first value will be used for calculating the mask.**

    ``rx``, ``ry``, ``darkstr``, ``brightstr``, ``lowsens``, ``highsens``, ``ss`` are all
    configurable per plane and iteration. `tuple` means iteration, `list` plane.

    `rx=(2.0, [2.0, 2.4], [2.2, 2.0, 2.1])` means three iterations.
     * 1st => 2.0 for all planes
     * 2nd => 2.0 for luma, 2.4 for chroma
     * 3rd => 2.2 for luma, 2.0 for u, 2.1 for v


    :param clip:                Source clip.
    :param rx:                  Horizontal radius for halo removal.
    :param ry:                  Vertical radius for halo removal.
    :param darkstr:             Strength factor for dark halos.
    :param brightstr:           Strength factor for bright halos.
    :param lowsens:             Sensitivity setting for defining how weak the dehalo has to be to get fully accepted.
    :param highsens:            Sensitivity setting for define how strong the dehalo has to be to get fully discarded.
    :param thmi:                Minimum threshold for sharp edges; keep only the sharpest edges (line edges).
    :param thma:                Maximum threshold for sharp edges; keep only the sharpest edges (line edges).
    :param thlimi:              Minimum limiting threshold; includes more edges than previously, but ignores simple details.
    :param thlima:              Maximum limiting threshold; includes more edges than previously, but ignores simple details.
    :param sigma_mask:          Blurring strength for the mask.
    :param ss:                  Supersampling factor, to avoid creation of aliasing.
    :param contra:              Contrasharpening. If True or int, will use :py:func:`contrasharpening`
                                otherwise uses :py:func:`contrasharpening_fine_dehalo` with specified level.
    :param exclude:             If True, add an addionnal step to exclude edges close to each other
    :param edgeproc:            If > 0, it will add the edgemask to the processing, defaults to 0.0
    :param edgemask:            Internal mask used for detecting the edges, defaults to Robinson3()
    :param planes:              Planes to process.
    :param show_mask:           Whether to show the computed halo mask. 1-7 values to select intermediate masks.
    :param mask_radius:         Mask expanding radius with ``gradient``.
    :param downscaler:          Scaler used to downscale the clip.
    :param upscaler:            Scaler used to upscale the downscaled clip.
    :param supersampler:        Scaler used to supersampler the rescaled clip to `ss` factor.
    :param supersampler_ref:    Reference scaler used to clamp the supersampled clip. Has to be blurrier.
    :param pre_ss:              Supersampling rate used before anything else.
    :param pre_supersampler:    Supersampler used for ``pre_ss``.
    :param pre_downscaler:      Downscaler used for undoing the upscaling done by ``pre_supersampler``.
    :param func:                Function from where this function was called.

    :return:                    Dehaloed clip.


    """
    assert clip.format

    func = func or fine_dehalo

    InvalidColorFamilyError.check(clip, (vs.GRAY, vs.YUV), func)

    if show_mask is not False and not (0 < int(show_mask) <= 7):
        raise CustomValueError('valid values for show_mask are 1–7!', func)

    thmif, thmaf, thlimif, thlimaf = [
        scale_value(x, 8, clip, ColorRange.FULL)
        for x in [thmi, thma, thlimi, thlima]
    ]

    peak = get_peak_value(clip)
    planes = normalize_planes(clip, planes)

    rx_i, ry_i = cround(to_arr(to_arr(rx)[0])[0]), cround(to_arr(to_arr(fallback(ry, rx))[0])[0])  # type: ignore

    work_clip, *chroma = split(clip) if planes == [0] else (clip, )

    # Main edges #
    # Basic edge detection, thresholding will be applied later.
    edges = edgemask.edgemask(work_clip)

    # Keeps only the sharpest edges (line edges)
    strong = norm_expr(edges, f'x {thmif} - {thmaf - thmif} / {peak} *', planes)

    # Extends them to include the potential halos
    large = expand(strong, rx_i, ry_i, planes=planes)

    # Exclusion zones #
    # When two edges are close from each other (both edges of a single
    # line or multiple parallel color bands), the halo removal
    # oversmoothes them or makes seriously bleed the bands, producing
    # annoying artifacts. Therefore we have to produce a mask to exclude
    # these zones from the halo removal.

    # Includes more edges than previously, but ignores simple details
    light = norm_expr(edges, f'x {thlimif} - {thlimaf - thlimif} / {peak} *', planes)

    # To build the exclusion zone, we make grow the edge mask, then shrink
    # it to its original shape. During the growing stage, close adjacent
    # edge masks will join and merge, forming a solid area, which will
    # remain solid even after the shrinking stage.
    # Mask growing
    shrink = expand(light, rx_i, ry_i, XxpandMode.ELLIPSE, planes=planes)

    # At this point, because the mask was made of a shades of grey, we may
    # end up with large areas of dark grey after shrinking. To avoid this,
    # we amplify and saturate the mask here (actually we could even
    # binarize it).
    shrink = norm_expr(shrink, 'x 4 *', planes)
    shrink = inpand(shrink, rx_i, rx_i, XxpandMode.ELLIPSE, planes=planes)

    # This mask is almost binary, which will produce distinct
    # discontinuities once applied. Then we have to smooth it.
    shrink = box_blur(shrink, 1, 2, planes)

    # Final mask building #

    # Previous mask may be a bit weak on the pure edge side, so we ensure
    # that the main edges are really excluded. We do not want them to be
    # smoothed by the halo removal.
    shr_med = combine([strong, shrink], ExprOp.MAX, planes=planes) if exclude else strong

    # Substracts masks and amplifies the difference to be sure we get 255
    # on the areas to be processed.
    mask = norm_expr([large, shr_med], 'x y - 2 *', planes)

    # If edge processing is required, adds the edgemask
    if edgeproc > 0:
        mask = norm_expr([mask, strong], f'x y {edgeproc} 0.66 * * +', planes)

    # Smooth again and amplify to grow the mask a bit, otherwise the halo
    # parts sticking to the edges could be missed.
    # Also clamp to legal ranges
    mask = box_blur(mask, planes=planes)

    mask = norm_expr(mask, f'x 2 * {ExprOp.clamp(0, peak)}', planes)

    # Masking #
    if show_mask:
        return [mask, shrink, edges, strong, light, large, shr_med][show_mask - 1]

    dehaloed = dehalo_alpha(
        work_clip, rx, ry, darkstr, brightstr, lowsens, highsens, sigma_mask, ss, planes, False, mask_radius,
        downscaler, upscaler, supersampler, supersampler_ref, pre_ss, pre_supersampler, pre_downscaler, func
    )

    if isinstance(contra, float):
        dehaloed = contrasharpening_dehalo(dehaloed, work_clip, contra, planes=planes)
    elif contra:
        dehaloed = contrasharpening(
            dehaloed, work_clip, None if contra is True else contra, planes=planes
        )

    y_merge = work_clip.std.MaskedMerge(dehaloed, mask, planes)

    if chroma:
        return join([y_merge, *chroma], clip.format.color_family)

    return y_merge


def fine_dehalo2(
    clip: vs.VideoNode, mode: ConvMode = ConvMode.SQUARE,
    radius: int = 2, dark: bool | None = True, planes: PlanesT = 0, show_mask: bool = False
) -> vs.VideoNode:
    """
    Halo removal function for 2nd order halos.

    :param clip:        Source clip.
    :param mode:        Horizontal/Vertical or both ways.
    :param radius:      Radius for mask growing.
    :param dark:        Whether to filter for dark or bright haloing.
                        None for disable merging with source clip.
    :param planes:      Planes to process.
    :param show_mask:   Whether to return the computed mask.

    :return:            Dehaloed clip.
    """
    assert clip.format

    if clip.format.color_family not in {vs.YUV, vs.GRAY}:
        raise ValueError('fine_dehalo2: format not supported')

    planes = normalize_planes(clip, planes)

    is_float = clip.format.sample_type == vs.FLOAT

    work_clip, *chroma = split(clip) if planes == [0] else (clip, )

    mask_h = mask_v = None

    # intended to be reversed
    if aka_expr_available:
        h_mexpr, v_mexpr = [
            ExprOp.convolution('x', coord, None, 4, False)
            for coord in [
                [[1, 2, 1], [0, 0, 0], [-1, -2, -1]],
                [[1, 0, -1], [2, 0, -2], [1, 0, -1]]
            ]
        ]

        if mode == ConvMode.SQUARE:
            do_mv = do_mh = True
        else:
            do_mv, do_mh = [
                mode == m for m in {ConvMode.HORIZONTAL, ConvMode.VERTICAL}
            ]

        mask_args = (h_mexpr, do_mv, do_mh, v_mexpr)

        mask_h, mask_v = [
            norm_expr(work_clip, [
                mexpr, 3, ExprOp.MUL, [omexpr, ExprOp.SUB] if do_om else None, ExprOp.clamp(0, 1) if is_float else None
            ], planes) if do_m else None
            for mexpr, do_m, do_om, omexpr in [mask_args, mask_args[::-1]]
        ]
    else:
        if mode in {ConvMode.SQUARE, ConvMode.VERTICAL}:
            mask_h = work_clip.std.Convolution([1, 2, 1, 0, 0, 0, -1, -2, -1], None, 4, planes, False)

        if mode in {ConvMode.SQUARE, ConvMode.HORIZONTAL}:
            mask_v = work_clip.std.Convolution([1, 0, -1, 2, 0, -2, 1, 0, -1], None, 4, planes, False)

        if mask_h and mask_v:
            mask_h = norm_expr([mask_h, mask_v], 'x 3 * y -', planes)
            mask_v = norm_expr([mask_v, mask_h], 'x 3 * y -', planes)
        elif mask_h:
            mask_h = norm_expr(mask_h, 'x 3 *', planes)
        elif mask_v:
            mask_v = norm_expr(mask_v, 'x 3 *', planes)

        if is_float:
            mask_h = mask_h and mask_h.std.Limiter(planes=planes)
            mask_v = mask_v and mask_v.std.Limiter(planes=planes)

    fix_h, fix_v = [
        norm_expr(work_clip, ExprOp.convolution('x', coord, mode=mode), planes)
        if aka_expr_available else
        work_clip.std.Convolution(coord, planes=planes, mode=mode)
        for coord, mode in [
            ([-1, -2, 0, 0, 40, 0, 0, -2, -1], ConvMode.HORIZONTAL),
            ([-2, -1, 0, 0, 40, 0, 0, -1, -2], ConvMode.VERTICAL)
        ]
    ]

    mask_h, mask_v = [
        masks.grow_mask(mask, radius, 1.8, planes, coordinates=coord) if mask else None
        for mask, coord in [
            (mask_h, [0, 1, 0, 0, 0, 0, 1, 0]), (mask_v, [0, 0, 0, 1, 1, 0, 0, 0])
        ]
    ]

    if is_float:
        mask_h = mask_h and mask_h.std.Limiter()
        mask_v = mask_v and mask_v.std.Limiter()

    if show_mask:
        if mask_h and mask_v:
            return combine([mask_h, mask_v], ExprOp.MAX)

        assert (ret_mask := mask_h or mask_v)

        return ret_mask

    dehaloed = work_clip
    op = '' if dark is None else ExprOp.MAX if dark else ExprOp.MIN

    if aka_expr_available and mask_h and mask_v and clip.format.sample_type is vs.FLOAT:
        dehaloed = norm_expr(
            [work_clip, fix_h, fix_v, mask_h, mask_v, clip], f'x 1 a - * y a * + 1 b - * z b * + c {op}', planes
        )
    else:
        for fix, mask in [(fix_h, mask_v), (fix_v, mask_h)]:
            if mask:
                dehaloed = dehaloed.std.MaskedMerge(fix, mask)

        if op:
            dehaloed = combine([work_clip, dehaloed], ExprOp(op))

    if not chroma:
        return dehaloed

    return join([dehaloed, *chroma], clip.format.color_family)


@disallow_variable_format
@disallow_variable_resolution
def dehalo_alpha(
    clip: vs.VideoNode, rx: FloatIterArr = 2.0, ry: FloatIterArr | None = None, darkstr: FloatIterArr = 0.0,
    brightstr: FloatIterArr = 1.0, lowsens: FloatIterArr = 50.0, highsens: FloatIterArr = 50.0,
    sigma_mask: float = 0.0, ss: FloatIterArr = 1.5, planes: PlanesT = 0, show_mask: bool = False,
    mask_radius: int = 1, downscaler: ScalerT = Mitchell, upscaler: ScalerT = BSpline,
    supersampler: ScalerT = Lanczos(3), supersampler_ref: ScalerT = Mitchell, pre_ss: float = 1.0,
    pre_supersampler: ScalerT = Nnedi3(0, field=0, shifter=NoShift), pre_downscaler: ScalerT = Point,
    func: FuncExceptT | None = None
) -> vs.VideoNode:
    """
    Reduce halo artifacts by nuking everything around edges (and also the edges actually).

    ``rx``, ``ry``, ``darkstr``, ``brightstr``, ``lowsens``, ``highsens``, ``ss`` are all
    configurable per plane and iteration. `tuple` means iteration, `list` plane.

    `rx=(2.0, [2.0, 2.4], [2.2, 2.0, 2.1])` means three iterations.
     * 1st => 2.0 for all planes
     * 2nd => 2.0 for luma, 2.4 for chroma
     * 3rd => 2.2 for luma, 2.0 for u, 2.1 for v

    :param clip:                Source clip.
    :param rx:                  Horizontal radius for halo removal.
    :param ry:                  Vertical radius for halo removal.
    :param darkstr:             Strength factor for dark halos.
    :param brightstr:           Strength factor for bright halos.
    :param lowsens:             Sensitivity setting for defining how weak the dehalo has to be to get fully accepted.
    :param highsens:            Sensitivity setting for define how strong the dehalo has to be to get fully discarded.
    :param sigma_mask:          Blurring strength for the mask.
    :param ss:                  Supersampling factor, to avoid creation of aliasing.
    :param planes:              Planes to process.
    :param show_mask:           Whether to show the computed halo mask.
    :param mask_radius:         Mask expanding radius with ``gradient``.
    :param downscaler:          Scaler used to downscale the clip.
    :param upscaler:            Scaler used to upscale the downscaled clip.
    :param supersampler:        Scaler used to supersampler the rescaled clip to `ss` factor.
    :param supersampler_ref:    Reference scaler used to clamp the supersampled clip. Has to be blurrier.
    :param pre_ss:              Supersampling rate used before anything else.
    :param pre_supersampler:    Supersampler used for ``pre_ss``.
    :param pre_downscaler:      Downscaler used for undoing the upscaling done by ``pre_supersampler``.
    :param func:                Function from where this function was called.

    :return:                    Dehaloed clip.
    """

    func = func or dehalo_alpha

    assert check_variable(clip, func)

    InvalidColorFamilyError.check(clip, (vs.GRAY, vs.YUV), func)

    peak = get_peak_value(clip)
    planes = normalize_planes(clip, planes)

    downscaler = Scaler.ensure_obj(downscaler, func)
    upscaler = Scaler.ensure_obj(upscaler, func)
    supersampler = Scaler.ensure_obj(supersampler, func)
    supersampler_ref = Scaler.ensure_obj(supersampler_ref, func)
    pre_supersampler = Scaler.ensure_obj(pre_supersampler, func)
    pre_downscaler = Scaler.ensure_obj(pre_downscaler, func)

    ry = fallback(ry, rx)  # type: ignore

    work_clip, *chroma = split(clip) if planes == [0] else (clip, )

    if pre_ss > 1.0:
        work_clip = pre_supersampler.scale(
            work_clip, mod4(work_clip.width * pre_ss), mod4(work_clip.height * pre_ss)
        )

    assert work_clip.format

    def _rescale(clip: vs.VideoNode, rx: float, ry: float) -> vs.VideoNode:
        return upscaler.scale(downscaler.scale(  # type: ignore
            clip, mod4(clip.width / rx), mod4(clip.height / ry)
        ), clip.width, clip.height)

    def _supersample(work_clip: vs.VideoNode, dehalo: vs.VideoNode, ss: float) -> vs.VideoNode:
        if ss <= 1.0:
            return repair(work_clip, dehalo, norm_rmode_planes(work_clip, 1, planes))

        w, h = mod4(work_clip.width * ss), mod4(work_clip.height * ss)
        ss_clip = norm_expr([
            supersampler.scale(work_clip, w, h),  # type: ignore
            supersampler_ref.scale(dehalo.std.Maximum(), w, h),  # type: ignore
            supersampler_ref.scale(dehalo.std.Minimum(), w, h)  # type: ignore
        ], 'x y min z max', planes)

        return supersampler.scale(ss_clip, work_clip.width, work_clip.height)  # type: ignore

    values = [rx, ry, darkstr, brightstr, lowsens, highsens, ss]

    iterations = max([(len(x) if isinstance(x, tuple) else 1) for x in values])

    values_norm: list[tuple[list[float], ...]] = zip(*[  # type: ignore[assignment]
        tuple(normalize_seq(x) for x in y)
        for y in [
            (*x, *((x[-1], ) * (len(x) - iterations))) if isinstance(x, tuple) else ((x, ) * iterations)
            for x in values
        ]
    ])

    for rx_i, ry_i, darkstr_i, brightstr_i, lowsens_i, highsens_i, ss_i in values_norm:
        if not all(x >= 1 for x in (*ss_i, *rx_i, *ry_i)):
            raise CustomIndexError('ss, rx, and ry must all be bigger than 1.0!', func)

        if not all(0 <= x <= 1 for x in (*brightstr_i, *darkstr_i)):
            raise CustomIndexError('brightstr, darkstr must be between 0.0 and 1.0!', func)

        if not all(0 <= x <= 100 for x in (*lowsens_i, *highsens_i)):
            raise CustomIndexError('lowsens and highsens must be between 0 and 100!', func)

        if len(set(rx_i)) == len(set(ry_i)) == 1 or planes == [0] or work_clip.format.num_planes == 1:
            dehalo = _rescale(work_clip, rx_i[0], ry_i[0])
        else:
            dehalo = join([
                _rescale(plane, rxp, ryp)
                for plane, rxp, ryp in zip(split(work_clip), rx_i, ry_i)
            ])

        mask = norm_expr(
            [masks.gradient(work_clip, mask_radius, planes), masks.gradient(dehalo, mask_radius, planes)],
            'x 0 = 1.0 x y - x / ? {lowsens} - x {peak} / 256 255 / + 512 255 / / {highsens} + * '
            '0.0 max 1.0 min {peak} *', planes, peak=peak,
            lowsens=[lo / 255 for lo in lowsens_i], highsens=[hi / 100 for hi in highsens_i]
        )

        conv_values = [float((sig_mask := bool(sigma_mask)))] * 9
        conv_values[5] = 1 / clamp(sigma_mask, 0, 1) if sig_mask else 1

        mask = mask.std.Convolution(conv_values, planes=planes)

        if show_mask:
            return mask

        dehalo = dehalo.std.MaskedMerge(work_clip, mask, planes=planes)

        if len(set(ss_i)) == 1 or planes == [0] or work_clip.format.num_planes == 1:
            dehalo = _supersample(work_clip, dehalo, ss_i[0])
        else:
            dehalo = join([
                _supersample(wplane, dplane, ssp)
                for wplane, dplane, ssp in zip(split(work_clip), split(dehalo), ss_i)
            ])

        dehalo = norm_expr(
            [work_clip, dehalo], 'x y < x x y - {darkstr} * - x x y - {brightstr} * - ?', planes,
            darkstr=darkstr_i, brightstr=brightstr_i
        )

        work_clip = dehalo

    if (dehalo.width, dehalo.height) != (clip.width, clip.height):
        dehalo = pre_downscaler.scale(work_clip, clip.width, clip.height)

    if not chroma:
        return dehalo

    return join([dehalo, *chroma], clip.format.color_family)
